#!/usr/bin/env python
# -*- python -*-
# $Id: rdfgen.py,v 1.28 2004/03/15 14:01:47 torh Exp $

true, false = 1, 0
DEBUG = true
error = 'rdfgen.error'

import os, sys, string, urlparse, getopt, re
from swuplib.driver import pkgdriver
from swuplib import config
from swuplib import utils
from swuplib import download
from swuplib import spiparser
from swuplib import gpg
from swuplib import help
import rdf 



SHORT_SWITCHES = "hqvSco:s:p:P:g:k:l:r:m:n:f:t:u:i:I:"
LONG_SWITCHES = ["help", 
                 "version", 
                 "quiet", 
                 "sign",
                 "compress",
                 "keyid", 
                 "password", 
                 "gpghomedir"]

# Defaults
RDFDIR = "rdfs"
RESOURCEDIRNAME = "resources"
PACKAGEDIRNAME = "packages"
PACKAGEPREFIX = ""
SIGPREFIX = ""
LATESTFILENAME = "latest.rdf"
RESOURCELIST_FILENAME = "resourcelist.rdf"
FILELISTNAME = "filelist"
LINK_RESOURCELISTS = []
LINK_FILELISTS = []
PGPHOMEDIR = "/root/.gnupg"
PGPFILE= "/root/pass"
PGPID = ""
SIGN=false
COMPRESS=false

# exceptions
class UpgradeException(Exception):
    def __init__(self, value=None):
        self.value = value
    def __str__(self):
        return str(self.value)

class DownloadException(UpgradeException): pass
class ResolveException(UpgradeException): pass
class FlushCacheException(UpgradeException): pass
class ParseException(UpgradeException): pass
class AuthException(DownloadException): pass

def escape_capabilityname(capabilityname):
    # Hack to escape / in requirements.
    # Example: /bin/bash -> #bin#bash
    capabilityname = string.replace(capabilityname, "_", "__")
    capabilityname = string.replace(capabilityname, "/", "_")
    return capabilityname


class rdfgen:
    bracket_regexp = re.compile('^\[.*\]$')

    def __init__(self, rdfdir, resourcedirname, packagedirname,
                 packageprefix, sigprefix, latestfilename,
                 resourcelist_filename, filelistname,
                 link_resourcelists, link_filelists, quiet):
        self.rdfdir = rdfdir
        self.resourcedirname = resourcedirname
        self.packagedirname = packagedirname
        self.packageprefix = packageprefix
        self.sigprefix = sigprefix
        self.latestfilename = latestfilename
        self.resourcelist_filename = resourcelist_filename
        self.filelistname = filelistname
        self.link_resourcelists = link_resourcelists
        self.link_filelists = link_filelists
        self.filelists = {}
        self.files = {}
        self.capabilities = {}
        self.resources = {}
        self.packages = {}
        self.latest = {}
        self.requirements = {}
        self.writer = rdf.writer()
        self.reader = spiparser.rdf_reader()
	self.quiet = quiet
	self._init_config()
        self._init_downloader()
        self.driver = pkgdriver.pkgdriver(self.config)

    def _init_config(self):
        self.config = config.parse()	
	    
    def _init_downloader(self):
        "Initialize downloader."
        try:
            self._downloader = download.download(PGPHOMEDIR,
			    None,
			    self.config["cachedir"],
			    self.config["tmpdir"])
        except download.DownloadException, err:
            raise UpgradeException, err
        except:
            raise

    def download_file(self, uri, siguri=None, CHECKSIG=1, TYPE=1):
        "Download file from uri and check signature. Returns local filename."
        "TYPE is: 1 = rpm-related, 2 = rdf-related"
        try:
            filename = self._downloader.download(uri, siguri, CHECKSIG, TYPE)
            return filename
        except download.AuthException, errmsg:
            raise AuthException, errmsg
        except download.DownloadException, errmsg:
            raise DownloadException, errmsg
        except:
            raise


    def get_filelists(self):
        files = {}
        for uri in self.link_filelists:
            protocol, server, path = urlparse.urlparse(uri)[:3]
            rdfdir = os.path.dirname(path)
            if not self.quiet: sys.stdout.write('Retrieving file list: %s\n' % uri)
            file = self.download_file(uri, TYPE=2)
            fd = open(file)
            lines = fd.readlines()
            fd.close()
            baseuri = os.path.dirname(uri)
            if not self.bracket_regexp.match(lines[0]):
                raise error, 'Wrong format in filelist: %s' % uri
            for line in lines:
                if self.bracket_regexp.match(line):
                    pkg = string.replace(line, '[', ']')
                    pkg = string.replace(pkg, ']', '')
                    pkg = os.path.join(rdfdir, pkg)
                    pkg = urlparse.urlunparse((protocol, server, pkg,
                                               '', '', ''))
                    files[pkg] = []
                else:
                    files[pkg].append(string.strip(line))
        return files


    def get_resourcelists(self):
        resources = {}
        for uri in self.link_resourcelists:
            protocol, server, path = urlparse.urlparse(uri)[:3]
            baseuri = os.path.dirname(uri)
            if not server:
                baseuri = os.path.join('../..', baseuri)
	    if not self.quiet: sys.stdout.write('Retrieving resource list: %s\n' % uri)
            retr_file = self.download_file(uri,TYPE=2)            
            newresources = self.reader.read_resourcelist(retr_file, baseuri)
            newresources.update(resources)
            resources = newresources
        return resources

        
    def prepare_package(self, filename):
        pkgdata = self.driver.get_package(filename, True)
        name = pkgdata['name']
        basename = pkgdata['basename']
        self.filelists[basename] = pkgdata['filelist']            
        
        # Adding the packageinfo to a dictionary with basename as key.
        if self.packages.has_key(basename):
            raise error, 'Duplicate packages found: %s' % basename
        self.packages[basename] = pkgdata

        # Adding a list of latest version packages with name as key.
        if self.latest.has_key(name):
            old = self.latest[name]
            if self.driver.version_compare(old, pkgdata) < 0:
                self.latest[name] = pkgdata
        else:
            self.latest[name] = pkgdata
            
        # Adding the packageinfo to the capability dictionary with
        # package name as key.
        if not self.capabilities.has_key(name):
            self.capabilities[name] = []
        self.capabilities[name].append(pkgdata)

        # Adding the packageinfo to the capability dictionary with the
        # "provide"-string as key.
        provides = pkgdata['provides']
        if provides:
            for capability in  provides:
                if not self.capabilities.has_key(capability):
                    self.capabilities[capability] = []
                self.capabilities[capability].append(pkgdata)

        # Adding the packageinfo to a dictionary of requirements with
        # capability (requirement) as key.
        requires = pkgdata['requires']
        if requires:
            for capability in requires:
                if not self.requirements.has_key(capability):
                    self.requirements[capability] = []
                self.requirements[capability].append(pkgdata)

    def generate_packages(self):
        package_formatter = rdf.package_formatter(\
            self.writer, self.resources, self.resourcedirname,
            self.packagedirname, self.packageprefix, self.sigprefix)

        # Write rdf-file for package here.
        for basename in self.packages.keys():
            if not self.quiet: sys.stdout.write('Writing rdf for package: %s\n' % basename)
            pkgdata = self.packages[basename]
            outfilename = os.path.join(self.rdfdir, self.packagedirname,
                                       (basename + ".rdf"))
            outfile = open(outfilename, "w")        
            package_formatter.set_fd(outfile)
            package_formatter.reset()
            package_formatter.feed(pkgdata)
            outfile.close()
	    #Sign file
            if SIGN and not PGPID == "":
                gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, outfilename)



    def generate_resources(self):
        resource_formatter = rdf.resource_formatter(self.writer,
                                                    self.packagedirname)
        resourcelist_formatter = rdf.resourcelist_formatter(\
            self.writer, self.resourcedirname)
        capfilenames = {}
        for capabilityname in self.capabilities.keys():
            if not self.quiet: sys.stdout.write('Writing rdf for resouce: %s\n' % capabilityname)
            capabilitydata = self.capabilities[capabilityname]
            escaped_capabilityname = rdf.escape_capabilityname(\
                capabilityname)
            capfilename = os.path.join(self.resourcedirname,
                                       (escaped_capabilityname + ".rdf"))
            outfilename = os.path.join(self.rdfdir, capfilename)
            capfilenames[capabilityname] = capfilename
            outfile = open(outfilename, "w")
            resource_formatter.set_fd(outfile)
            resource_formatter.reset()
            resource_formatter.feed(capabilityname, capabilitydata)
            outfile.close()
            #Sign file
            if SIGN and not PGPID == "":
                gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, outfilename)

        
        outfilename = os.path.join(self.rdfdir, self.resourcelist_filename)
        outfile = open(outfilename, "w")
        resourcelist_formatter.set_fd(outfile)
        resourcelist_formatter.reset()
        resourcelist_formatter.feed(capfilenames)
        outfile.close()
        if COMPRESS:
            utils.compress_file(outfilename)
        #Sign file
        if SIGN and not PGPID == "":
            gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, outfilename)
        

    def generate_latestlist(self):
        packagelist_formatter = rdf.packagelist_formatter(\
            self.writer, self.packagedirname, self.resourcelist_filename)

        outfilename = os.path.join(self.rdfdir, self.latestfilename)
        outfile = open(outfilename, "w")

        packagelist_formatter.set_fd(outfile)
        packagelist_formatter.reset()
        packagelist_formatter.feed(self.latest)
        outfile.close()
        #Md5
        utils.md5sumtofile(outfilename)
        #Compress
        if COMPRESS:
            utils.compress_file(outfilename)
	#Sign file
        if SIGN and not PGPID == "":
            gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, outfilename)

    def generate_filelist(self):
        filelist_formatter = rdf.filelist_formatter(self.packagedirname)
        outfilename = os.path.join(self.rdfdir, self.filelistname)
        outfile = open(outfilename, "w")
        filelist_formatter.set_fd(outfile)
        filelist_formatter.reset()
        filelist_formatter.feed(self.filelists)
        outfile.close()
        #Md5
        utils.md5sumtofile(outfilename)
        if COMPRESS:
            utils.compress_file(outfilename)
	#Sign file
        if SIGN and not PGPID == "":
            gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, outfilename)

    def get_resourcefilename(self, capname):
        return os.path.join(self.resourcedirname,
                            escape_capabilityname(capname)+'.rdf')

    def generate_rdfs(self, filenames):
        outputdir = self.rdfdir

        for filename in filenames:
            if not self.quiet: sys.stdout.write('Reading headers and signing: %s\n' % filename)
            self.prepare_package(filename)
	    #Sign file
            if SIGN and not PGPID == "":
                gpg.gpg_signfile(PGPFILE, PGPID, PGPHOMEDIR, filename)

        # Checking if there are any requirements on capabilities that are not
        # found in the capabilities dictionary. If so, look for the
        # capability in the file lists, and add the capability to the
        # capability dictionary with the packageinfo for the package in
        # which the file was found.
        incl_resources = None
        incl_files = None

        # Update resources from capabilities.
        for capname in self.capabilities.keys():
            self.resources[capname] = os.path.join(\
                '..', self.get_resourcefilename(capname))

	errors = ''
        for requirement in self.requirements.keys():
            if not self.resources.has_key(requirement):
                # Search for file requires in other packages
                for basename in self.packages.keys():
                    filelist = self.filelists[basename]
                    if requirement in filelist:
                        packageinfo = self.packages[basename]
                        if not self.capabilities.has_key(requirement):
                            self.capabilities[requirement] = []
                        self.capabilities[requirement].append(packageinfo)
                        self.resources[requirement] = \
                            os.path.join(\
                            '..', self.get_resourcefilename(requirement))

            # Now search included resourcelists.
            if not incl_resources:
                incl_resources = self.get_resourcelists()
            if requirement in incl_resources.keys():
                uri = incl_resources[requirement]
                protocol, server, path = urlparse.urlparse(uri)[:3]
                if not (protocol and server):
                   path = os.path.join('../..', path)
                   baseuri = os.path.dirname(path)
                else:
                   baseuri = os.path.dirname(uri)
                filename = self.download_file(uri)
                    
                resourceinfo = self.reader.read_resource(\
                    filename, baseuri)
                    
                for index in range(len(resourceinfo)):
                    resource = resourceinfo[index]
                    
                    uri = resource['uri']
                    baseuri = os.path.dirname(uri)
                    filename = self.download_file(uri)
                    packageinfo = self.reader.read_package_short(\
                        filename, baseuri)
                    packageinfo['uri'] = uri
                    if not self.capabilities.has_key(requirement):
                        self.capabilities[requirement] = []
                    self.capabilities[requirement].append(packageinfo) 
                    self.resources[requirement] = os.path.join(\
                        '..', self.get_resourcefilename(requirement))
            # Now search included filelists.
            if not incl_files:
                incl_files = self.get_filelists()
            for uri in incl_files.keys():
                if requirement in incl_files[uri]:
                    # We found the file. It is in package with package
                    # description found at <uri>. Generate a resource
                    # for this file locally.

                    # FIXME: what if more packages within the same distro
                    # supply the same file?
                    protocol, server, path = urlparse.urlparse(uri)[:3]
                    if not (protocol and server):
                        path = os.path.join('../..', path)
                        baseuri = os.path.dirname(path)
                    else:
                        baseuri = os.path.dirname(uri)
                    filename = self.download_file(uri)
                    
                    packageinfo = self.reader.read_package_short(\
                        filename, baseuri)
                    packageinfo['uri'] = uri
                    if not self.capabilities.has_key(requirement):
                        self.capabilities[requirement] = []
                    self.capabilities[requirement].append(packageinfo) 
                    self.resources[requirement] = os.path.join(\
                        '..', self.get_resourcefilename(requirement))
                       
            if self.resources.has_key(requirement):
                continue
            if not self.capabilities.has_key(requirement):
                requiringPackages = []
                for package in self.requirements[requirement]:
                    requiringPackages.append(package['name'])
		errors += "%s is missing the resource %s\n" \
			% (requiringPackages,requirement)

	if errors:
		sys.stderr.write("%s\n" %errors)
                
        # Write a one-file package listing with name, version, release and
        # resource-pointer to rdf for package.
        # Send whole dict of packages
        self.generate_latestlist()

        # Write rdf-file for resources/capabilities here.
        self.generate_resources()

        # Write package rdfs.
        self.generate_packages()

        # Write filelist.
        self.generate_filelist()


    def rdfify_files(self, of_files):
        to_directory = os.getcwd()
        resourcedir = os.path.join(self.rdfdir,self.resourcedirname)
        packagedir = os.path.join(self.rdfdir,self.packagedirname)
        if not os.path.isdir(resourcedir):
            os.makedirs(resourcedir)
        if not os.path.isdir(packagedir):
            os.makedirs(packagedir)
        for i in range( len(of_files)):
            of_files[i] = string.strip( of_files[i] )
        self.generate_rdfs( of_files )


    def main(self, of):
	realfilelist = []
        for fileordir in of:
            if os.path.isdir(fileordir):
               filelist = os.listdir(fileordir)
               for file in filelist:
                   realfile = os.path.join(fileordir,file)
                   if os.path.isfile(realfile) and \
                       os.path.splitext(realfile)[1].lower() == ".rpm":
                       realfilelist.append(realfile)
            elif os.path.isfile(fileordir) and \
                       os.path.splitext(fileordir)[1].lower() == ".rpm":
                       realfilelist.append(fileordir)
#		raise error, "No such file %s" % file
        self.rdfify_files(realfilelist)





def generate(files, rdfdir=RDFDIR, resourcedirname=RESOURCEDIRNAME,
             packagedirname=PACKAGEDIRNAME, packageprefix=PACKAGEPREFIX,
             sigprefix=SIGPREFIX, latestfilename=LATESTFILENAME,
             resourcelist_filename=RESOURCELIST_FILENAME,
             filelistname=FILELISTNAME,
             link_resourcelists=LINK_RESOURCELISTS,
             link_filelists=LINK_FILELISTS,
	     quiet=0):
    
    generator = rdfgen(rdfdir, resourcedirname, packagedirname,
                       packageprefix, sigprefix, latestfilename,
                       resourcelist_filename, filelistname,
                       link_resourcelists, link_filelists, quiet)
    generator.main(files)



def main(args):
    try:
        opts, args = getopt.getopt(args, SHORT_SWITCHES, LONG_SWITCHES)
        optsdict = {}
        for (opt,val) in opts:
            optsdict[opt] = val
        switches = optsdict.keys()
    except getopt.GetoptError, msg:
        sys.stderr.write(str(msg) + "\nUse -h for help.\n")
        sys.exit(1)
    except:
        raise

    if (not switches) and (not args):
        help.helpme(false)
        sys.exit(1)
    
    if ('-h' in switches) or ('--help' in switches):
        help.helpme(false)
        sys.exit(0)

    if ('-v' in switches) or ('--version' in switches):
        help.version(false)
        sys.exit(0)

    if '-f' in switches:
        fd = open(args[2], "r")
        files = fd.readlines()
        fd.close()
    else:
        files = args

    if '-o' in switches:
        rdfdir = optsdict['-o']
    else:
        rdfdir = RDFDIR

    if '-s' in switches:
        sigprefix = optsdict['-s']
    else:
        sigprefix = SIGPREFIX

    if '-p' in switches:
        packageprefix = optsdict['-p']
    else:
        packageprefix = PACKAGEPREFIX

    if '-l' in switches:
        latestfilename = optsdict['-l']
    else:
        latestfilename = LATESTFILENAME

    if '-r' in switches:
        resourcelist_filename = optsdict['-r']
    else:
        resourcelist_filename = RESOURCELIST_FILENAME

    if '-n' in switches:
        resourcedirname = optsdict['-n']
    else:
        resourcedirname = RESOURCEDIRNAME

    if '-m' in switches:
        packagedirname = optsdict['-m']
    else:
        packagedirname = PACKAGEDIRNAME

    if '-t' in switches:
        filelistname = optsdict['-t']
    else:
        filelistname = FILELISTNAME

    if '-i' in switches:
        link_resourcelists = []
        for (opt, optval) in opts:
            if opt == "-i":
                link_resourcelists.append(optval)
    else:
        link_resourcelists = []

    if '-I' in switches:
        link_filelists = []
        for (opt, optval) in opts:
            if opt == "-I":
                link_filelists.append(optval)
    else:
        link_filelists = []

    if '-S' in switches or '--sign' in switches:
        globals()['SIGN']=true

    if '-c' in switches or '--compress' in switches:
        globals()['COMPRESS']=true

    if '-k' in switches:
        globals()['PGPID'] = optsdict['-k']
        
    if '-P' in switches:
        globals()['PGPFILE'] = optsdict['-P']

    if '-g' in switches:
        globals()['PGPHOMEDIR'] = optsdict['-g']

    if '-q' in switches or '--quiet' in switches:
	quiet = 1
    else:
	quiet = 0

    
    try:
        generate(files, rdfdir, resourcedirname, packagedirname,
                 packageprefix, sigprefix, latestfilename,
                 resourcelist_filename, filelistname,
                 link_resourcelists, link_filelists, quiet)
    except error, errmsg:
        sys.stderr.write("%s\n\n" %errmsg)
        sys.exit(1)
    except:
        raise
    
    
if __name__ == '__main__':
    main(sys.argv)



